from ast import literal_eval
import logging
import logging.config
import os
import time
from influxdb_client import InfluxDBClient
from influxdb_client.client.write_api import SYNCHRONOUS
from influxdb_client.client.write.point import Point
from influxdb_client.domain.write_precision import WritePrecision
from typing import List, Dict

from .. import settings

"""
Using these functions would look like:

from utils.instrument import Instrumentation
stats = Instrumentation()
stats.write_stats([{"metric": "objects", "fields"{ "scraped": 10}, "tags": {"jurisdiction": "ca"}])
stats.close()
"""


class Instrumentation(object):
    def __init__(self) -> None:
        """
        We're currently leveraging https://github.com/civic-eagle/statsd-http-proxy
        to have an easy to reason about stats aggregation tool
        on the other end of an internet connection

        Essentially statsd-style metrics, we can rely on the aggregator
        to continually emit these stats so we get useful stats for monitoring/reporting
        """
        logging.config.dictConfig(settings.LOGGING)
        self.logger = logging.getLogger("openstates.stats")
        # use a literal_eval to properly turn a string into a bool (literal_eval 'cause it's safer than stdlib eval)
        self.enabled = literal_eval(os.environ.get("STATS_ENABLED", "False"))
        if not self.enabled:
            self.logger.debug("Stat emission is not enabled.")
            return
        token: str = os.environ.get("OPENSTATES_STATS_AUTH_TOKEN", "")
        self._batch: List[Dict] = list()
        self.prefix: str = os.environ.get("STATS_PREFIX", "openstates_")
        self.endpoint: str = os.environ.get("STATS_ENDPOINT", "")
        self.bucket: str = os.environ.get("STATS_BUCKET", "openstates")
        if self.endpoint.endswith("/"):
            self.endpoint = self.endpoint.strip("/")
        client = InfluxDBClient(
            url=self.endpoint,
            token=token,
            org="openstates",
            enable_gzip=True,
        )
        self.write_api = client.write_api(write_options=SYNCHRONOUS)
        self.batch_size: int = int(os.environ.get("STATS_BATCH_SIZE", 50))
        self.logger.debug(
            f"Stats emission to {self.endpoint} configured with batch size: {self.batch_size}"
        )

    def _send_stats(self, force: bool = False) -> None:
        """
        Needs to be broken out from _process_metric to have a
        "write at shutdown" function (force=True)
        Even with force=True, we should only write when there is data
        to write. Otherwise, just skip things.
        """
        batch_len = len(self._batch)
        if (force and batch_len > 0) or batch_len > self.batch_size:
            if not self.enabled:
                self.logger.debug("Stats diasbled. Skipping send")
                return
            if not self.endpoint:
                self.logger.debug("No stats endpoint defined. Not emitting stats")
                return
            points = []
            for m in self._batch:
                if self.prefix:
                    p = Point(f"{self.prefix}{m['metric']}")
                else:
                    p = Point(m["metric"])
                p.time(m["timestamp"], WritePrecision.S)
                """
                use list comprehensions 'cause they're technically faster than for loops
                But this is simply turning a dictionary of k=v pairs into tags/fields
                in the point object
                """
                [p.tag(t, v) for t, v in m.get("tags", {}).items()]
                [p.field(f, v) for f, v in m["fields"].items()]
                points.append(p)
            self.logger.debug(f"Sending stats batch: {self._batch}")
            try:
                self.write_api.write(self.bucket, record=points)
                self._batch = list()
            except Exception as e:
                self.logger.warning(
                    f"Failed to write {len(self._batch)} stats to {self.endpoint} :: {e}"
                )

    def write_stats(
        self,
        metrics: List[Dict],
    ) -> None:
        """
        Ensure consistent formatting of data objects to add to batch for sending
        Primarily, we'll force a timestamp on every metric in a consistent manner
        """
        if not self.enabled:
            return
        ts = int(time.time())
        for m in metrics:
            m["timestamp"] = ts
            self._batch.append(m)
        """
        we attempt to send (without forcing) after
        adding each stat to make sure we emit
        batches as quickly as we can, without
        overloading the write endpoint with tiny writes
        """
        self._send_stats()

    def close(self) -> None:
        """
        "Shut down" our instrumentation connection
        Currently just forces any batched stats out
        Keep as a wrapper script for later extensibility
        """
        if not self.enabled:
            return
        self._send_stats(force=True)



from . import states
from .pr import PR
from .dc import DC
from .us import US

STATES = states.STATES
TERRITORIES = states.TERRITORIES
STATES_AND_TERRITORIES = STATES + TERRITORIES + states.DISTRICTS
STATES_AND_DC_PR = STATES + [DC, PR]
INTERNATIONAL = states.INTERNATIONAL + [US]

STATES_BY_ABBR = {s.abbr: s for s in STATES_AND_TERRITORIES + INTERNATIONAL}
STATES_BY_JID = {s.jurisdiction_id: s for s in STATES_AND_TERRITORIES + INTERNATIONAL}
STATES_BY_NAME = {s.name.lower(): s for s in STATES_AND_TERRITORIES + INTERNATIONAL}

NON_US_INTERNATIONAL_ABBRS = [k.abbr for k in INTERNATIONAL if k.abbr != "US"]
INTERNATIONAL_ABBRS = [k.abbr for k in INTERNATIONAL]



# Generated by Django 3.2.2 on 2021-11-08 20:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "data",
            "0043_legislativesession_active",
        ),
    ]

    operations = [
        migrations.AddField(
            model_name="bill",
            name="citations",
            field=models.JSONField(blank=True, default=dict),
        ),
    ]



from .base import BaseModel, Scraper
from .popolo import Organization
from .schemas.jurisdiction import schema
from ..metadata import lookup


_name_fixes = {
    "SouthCarolina": "South Carolina",
    "NorthCarolina": "North Carolina",
    "SouthDakota": "South Dakota",
    "NorthDakota": "North Dakota",
    "RhodeIsland": "Rhode Island",
    "NewHampshire": "New Hampshire",
    "NewJersey": "New Jersey",
    "NewYork": "New York",
    "NewMexico": "New Mexico",
    "WestVirginia": "West Virginia",
    "PuertoRico": "Puerto Rico",
    "DistrictOfColumbia": "District of Columbia",
    "UnitedStates": "United States",
    "southafrica": "South Africa",
    "VirginIslands": "Virgin Islands",
    "AmericanSamoa": "American Samoa",
    "NorthernMarianaIslands": "Northern Mariana Islands",
}


class State(BaseModel):
    """ Base class for a jurisdiction """

    _type = "jurisdiction"
    _schema = schema

    # schema objects
    legislative_sessions = []
    extras = {}

    # non-db properties
    scrapers = {}
    default_scrapers = None
    ignored_scraped_sessions = []
    _metadata = None

    def __init__(self):
        super(BaseModel, self).__init__()
        self._related = []
        self.extras = {}

    @property
    def classification(self):
        if any(c == self.name for c in ["United States", "South Africa"]):
            return "country"
        else:
            return "state"

    @property
    def metadata(self):
        if not self._metadata:
            name = _name_fixes.get(self.__class__.__name__, self.__class__.__name__)
            self._metadata = lookup(name=name)
        return self._metadata

    @property
    def division_id(self):
        return self.metadata.division_id

    @property
    def jurisdiction_id(self):
        return "{}/government".format(
            self.division_id.replace("ocd-division", "ocd-jurisdiction"),
        )

    @property
    def name(self):
        return self.metadata.name

    @property
    def url(self):
        return self.metadata.url

    def get_organizations(self):
        legislature = Organization(
            name=self.metadata.legislature_name, classification="legislature"
        )
        yield legislature
        if not self.metadata.unicameral:
            yield Organization(
                self.metadata.upper.name,
                classification="upper",
                parent_id=legislature._id,
            )
            yield Organization(
                self.metadata.lower.name,
                classification="lower",
                parent_id=legislature._id,
            )

    def get_session_list(self) -> list[str]:
        raise NotImplementedError()

    _id = jurisdiction_id

    def as_dict(self):
        return {
            "_id": self.jurisdiction_id,
            "id": self.jurisdiction_id,
            "name": self.name,
            "url": self.url,
            "division_id": self.division_id,
            "classification": self.classification,
            "legislative_sessions": self.legislative_sessions,
            "extras": self.extras,
        }

    def __str__(self):
        return self.name


class JurisdictionScraper(Scraper):
    def scrape(self):
        # yield a single Jurisdiction object
        yield self.jurisdiction

        # yield all organizations
        for org in self.jurisdiction.get_organizations():
            yield org



import datetime
from django.db import models
from django.db.models import Q, QuerySet
from .base import OCDBase, LinkBase, OCDIDField, RelatedBase, IdentifierBase
from .division import Division
from .jurisdiction import Jurisdiction
from .. import common
from ...utils import abbr_to_jid


# the actual models


class Organization(OCDBase):
    """
    A group of people, typically in a legislative or rule-making context.
    """

    id = OCDIDField(ocd_type="organization")
    name = models.CharField(max_length=300, help_text="The name of the Organization.")
    parent = models.ForeignKey(
        "self",
        related_name="children",
        null=True,
        # parent can be deleted w/o affecting children
        on_delete=models.SET_NULL,
        help_text="A link to another Organization that serves as this Organization's parent.",
    )
    jurisdiction = models.ForeignKey(
        Jurisdiction,
        related_name="organizations",
        null=True,
        # deletion of a jurisdiction should be hard
        on_delete=models.PROTECT,
        help_text="A link to the Jurisdiction that contains this Organization.",
    )
    classification = models.CharField(
        max_length=100,
        blank=True,
        choices=common.ORGANIZATION_CLASSIFICATION_CHOICES,
        help_text="The type of Organization being defined.",
    )
    sources = models.JSONField(default=list, blank=True)
    links = models.JSONField(default=list, blank=True)
    other_names = models.JSONField(default=list, blank=True)

    def __str__(self):
        return self.name

    # Access all "ancestor" organizations
    def get_parents(self):
        org = self
        while True:
            org = org.parent
            # Django accesses parents lazily, so have to check if one actually exists
            if org:
                yield org
            else:
                break

    def get_current_members(self):
        """return all Person objects w/ current memberships to org"""
        today = datetime.date.today().isoformat()

        return Person.objects.filter(
            Q(memberships__start_date="") | Q(memberships__start_date__lte=today),
            Q(memberships__end_date="") | Q(memberships__end_date__gte=today),
            memberships__organization_id=self.id,
        )

    class Meta:
        db_table = "opencivicdata_organization"
        index_together = [
            ["jurisdiction", "classification", "name"],
            ["classification", "name"],
        ]


class Post(OCDBase):
    """
    A position in an organization that exists independently of the person holding it.
    """

    id = OCDIDField(ocd_type="post")
    label = models.CharField(max_length=300, help_text="A label describing the Post.")
    role = models.CharField(
        max_length=300,
        blank=True,
        help_text="The function that the holder of the post fulfills.",
    )
    organization = models.ForeignKey(
        Organization,
        related_name="posts",
        help_text="The Organization in which the post is held.",
        on_delete=models.CASCADE,
    )
    division = models.ForeignKey(
        Division,
        related_name="posts",
        null=True,
        blank=True,
        default=None,
        help_text="The Division where the post exists.",
        # if the division goes away the post is just jurisdiction-less
        on_delete=models.SET_NULL,
    )
    maximum_memberships = models.PositiveIntegerField(
        default=1, help_text="The maximum number of people who can hold this Post."
    )

    class Meta:
        db_table = "opencivicdata_post"
        index_together = [["organization", "label"]]

    def __str__(self):
        return "{} - {} - {}".format(self.role, self.label, self.organization)


class PersonQuerySet(QuerySet):
    def member_of(self, organization_name, current_only=True, post=None):
        if organization_name.startswith("ocd-organization/"):
            org_filter = Q(memberships__organization_id=organization_name)
        else:
            org_filter = Q(memberships__organization__name=organization_name)

        qs = self

        if current_only:
            today = datetime.date.today().isoformat()
            org_filter &= (
                Q(memberships__start_date="") | Q(memberships__start_date__lte=today)
            ) & Q(memberships__end_date="") | Q(memberships__end_date__gte=today)
        if post:
            org_filter &= Q(memberships__post__label=post)
        return qs.filter(org_filter).distinct()

    def active(self):
        return self.exclude(current_role=None)

    def current_legislators_with_roles(self, chambers):
        today = datetime.date.today().isoformat()
        return self.filter(
            Q(memberships__organization__in=chambers),
            Q(memberships__start_date="") | Q(memberships__start_date__lte=today),
            Q(memberships__end_date="") | Q(memberships__end_date__gte=today),
        ).prefetch_related(
            "memberships", "memberships__organization", "memberships__post"
        )

    def search(self, query, *, state=None, current=True):
        if current:
            people = self.active().filter(name__icontains=query)
        else:
            people = self.filter(name__icontains=query)

        if state:
            people = people.filter(
                memberships__organization__jurisdiction_id=abbr_to_jid(state)
            )
        return people


class Person(OCDBase):
    """
    An individual that has served in a political office.
    """

    objects = PersonQuerySet.as_manager()

    id = OCDIDField(ocd_type="person")
    name = models.CharField(
        max_length=300, db_index=True, help_text="A Person's preferred full name."
    )
    family_name = models.CharField(
        max_length=100, blank=True, help_text="A Person's family name."
    )
    given_name = models.CharField(
        max_length=100, blank=True, help_text="A Person's given name."
    )
    image = models.URLField(
        blank=True,
        max_length=2000,
        help_text="A URL leading to an image that identifies the Person visually.",
    )
    email = models.CharField(
        max_length=300,
        help_text="The official email address of the Person.",
        blank=True,
        default="",
    )
    gender = models.CharField(max_length=100, blank=True, help_text="A Person's gender")
    biography = models.TextField(
        blank=True, help_text="An extended account of a Person's life."
    )
    birth_date = models.CharField(
        max_length=10,
        blank=True,
        help_text="The date of a Person's birth in YYYY[-MM[-DD]] string format.",
    )
    death_date = models.CharField(
        max_length=10,
        blank=True,
        help_text="The date of a Person's death in YYYY[-MM[-DD]] string format.",
    )

    # computed fields
    primary_party = models.CharField(
        max_length=100,
        default="",
        help_text="Primary party an individual is associated with.",
    )
    current_jurisdiction = models.ForeignKey(
        Jurisdiction,
        related_name="current_people",
        on_delete=models.PROTECT,
        null=True,
        default=None,
    )
    current_role = models.JSONField(null=True, default=None)

    def __str__(self):
        return self.name

    class Meta:
        db_table = "opencivicdata_person"
        verbose_name_plural = "people"

    def add_other_name(self, name, note=""):
        PersonName.objects.create(name=name, note=note, person_id=self.id)


class PersonIdentifier(IdentifierBase):
    """
    Upstream identifier for a Person.
    """

    person = models.ForeignKey(
        Person,
        related_name="identifiers",
        on_delete=models.CASCADE,
        help_text="A link to the Person connected to this alternative identifier.",
    )

    class Meta:
        db_table = "opencivicdata_personidentifier"


class PersonName(RelatedBase):
    """
    Alternate or former name of a Person.
    """

    name = models.CharField(
        max_length=500, db_index=True, help_text="An alternative name."
    )
    note = models.CharField(
        max_length=500,
        blank=True,
        help_text="A short, optional note about alternative name.",
    )
    start_date = models.CharField(
        max_length=10,
        blank=True,
        help_text="An optional start date for usage of the alternative name "
        "in YYYY[-MM[-DD]] string format.",
    )
    end_date = models.CharField(
        max_length=10,
        blank=True,
        help_text="An optional end date for usage of the alternative name in "
        "YYYY[-MM[-DD]] string format.",
    )
    person = models.ForeignKey(
        Person,
        related_name="other_names",
        on_delete=models.CASCADE,
        help_text="A link to the Person connected to this alternative name.",
    )

    class Meta:
        db_table = "opencivicdata_personname"

    def __str__(self):
        return "{} ({})".format(self.name, self.note)


OFFICE_CHOICES = (
    ("district", "District Office"),
    ("district-mail", "District Mailing Address"),
    ("capitol", "Capitol Office"),
    ("capitol-mail", "Capitol Mailing Address"),
    ("primary", "Primary Office"),
    ("home", "Home"),
)


class PersonOffice(RelatedBase):
    """
    Office for a Person.
    """

    classification = models.CharField(max_length=20, choices=OFFICE_CHOICES)
    address = models.CharField(max_length=300, blank=True, default="")
    voice = models.CharField(max_length=30, blank=True, default="")
    fax = models.CharField(max_length=30, blank=True, default="")
    name = models.CharField(max_length=200, blank=True, default="")

    @property
    def display_name(self):
        return self.name or self.get_classification_display()

    person = models.ForeignKey(Person, related_name="offices", on_delete=models.CASCADE)

    class Meta:
        db_table = "openstates_personoffice"

    def __str__(self):
        return f"{self.person} {self.display_name}"


class PersonLink(LinkBase):
    """
    URL for a document about a Person.
    """

    person = models.ForeignKey(
        Person,
        related_name="links",
        on_delete=models.CASCADE,
        help_text="A reference to the Person connected to this link.",
    )

    class Meta:
        db_table = "opencivicdata_personlink"


class PersonSource(LinkBase):
    """
    Source used in assembling a Person.
    """

    person = models.ForeignKey(
        Person,
        related_name="sources",
        on_delete=models.CASCADE,
        help_text="A link to the Person connected to this source.",
    )

    class Meta:
        db_table = "opencivicdata_personsource"


class Membership(OCDBase):
    """
    A relationship between a Person and an Organization, possibly including a Post.
    """

    id = OCDIDField(ocd_type="membership")
    organization = models.ForeignKey(
        Organization,
        related_name="memberships",
        # memberships will go away if the org does
        on_delete=models.CASCADE,
        help_text="A link to the Organization in which the Person is a member.",
    )
    person = models.ForeignKey(
        Person,
        related_name="memberships",
        null=True,
        # Membership will just unlink if the person goes away
        on_delete=models.SET_NULL,
        help_text="A link to the Person that is a member of the Organization.",
    )
    person_name = models.CharField(
        max_length=300,
        blank=True,
        default="",
        help_text="The name of the Person that is a member of the Organization.",
    )
    post = models.ForeignKey(
        Post,
        related_name="memberships",
        null=True,
        # Membership will just unlink if the post goes away
        on_delete=models.SET_NULL,
        help_text="	The Post held by the member in the Organization.",
    )
    role = models.CharField(
        max_length=300,
        blank=True,
        help_text="The role that the member fulfills in the Organization.",
    )
    start_date = models.CharField(
        max_length=10,
        blank=True,
        help_text="The date on which the relationship began in YYYY[-MM[-DD]] string format.",
    )
    end_date = models.CharField(
        max_length=10,
        blank=True,
        help_text="The date on which the relationship ended in YYYY[-MM[-DD]] string format.",
    )

    class Meta:
        db_table = "opencivicdata_membership"
        index_together = [["organization", "person", "post"]]

    def __str__(self):
        return "{} in {} ({})".format(self.person, self.organization, self.role)



from ..models import State, Chamber, simple_numbered_districts

AR = State(
    name="Arkansas",
    abbr="AR",
    capital="Little Rock",
    capital_tz="America/Chicago",
    fips="05",
    unicameral=False,
    legislature_name="Arkansas General Assembly",
    legislature_organization_id="ocd-organization/4d1533a8-8d01-41f1-8b2a-4f07e41b5ed2",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/24b198d5-a37c-577e-8812-d53456b6caaa",
    division_id="ocd-division/country:us/state:ar",
    jurisdiction_id="ocd-jurisdiction/country:us/state:ar/government",
    url="http://www.arkleg.state.ar.us",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/9e2c94af-8e67-438b-8a47-7eac48112403",
        num_seats=100,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:ar", "lower", 100
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/f80b9e71-0c25-4b8b-a87d-7266a3f84313",
        num_seats=35,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:ar", "upper", 35
        ),
    ),
)



from ..models import State, Chamber, simple_numbered_districts

GU = State(
    name="Guam",
    abbr="GU",
    capital="Hagåtña",
    capital_tz="Pacific/Guam",
    fips="66",
    url="https://guam.gov",
    unicameral=True,
    legislature_name="Guam Legislature",
    legislature_organization_id="ocd-organization/74999629-090b-4aa4-92ed-6eb86396017d",
    executive_name="Governor of Guam",
    executive_organization_id="ocd-organization/2ac004ab-9614-47a0-b9b4-58d5c90f8081",
    division_id="ocd-division/country:us/territory:gu",
    jurisdiction_id="ocd-jurisdiction/country:us/territory:gu/government",
    legislature=Chamber(
        chamber_type="unicameral",
        name="Guam Legislature",
        organization_id="ocd-organization/74999629-090b-4aa4-92ed-6eb86396017d",
        num_seats=15,
        title="Legislators",
        districts=simple_numbered_districts(
            "ocd-division/country:us/territory:gu", "legislature", 15
        ),
    ),
)



from ..models import State, Chamber, simple_numbered_districts

IN = State(
    name="Indiana",
    abbr="IN",
    capital="Indianapolis",
    capital_tz="America/Indiana/Indianapolis",
    fips="18",
    unicameral=False,
    legislature_name="Indiana General Assembly",
    legislature_organization_id="ocd-organization/7de4a037-2e98-46d0-975c-67d42eaff437",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/3ecd0c76-5e2c-4e7c-9882-615d64fd0f47",
    division_id="ocd-division/country:us/state:in",
    jurisdiction_id="ocd-jurisdiction/country:us/state:in/government",
    url="http://www.in.gov/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/10bc85d1-df30-4bb9-a8d0-e158225b4431",
        num_seats=100,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:in", "lower", 100
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/b8946e05-f621-442a-8352-cc51db92980c",
        num_seats=50,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:in", "upper", 50
        ),
    ),
)



from .. import metadata


def jid_to_abbr(jid: str) -> str:
    return metadata.lookup(jurisdiction_id=jid).abbr.lower()


def abbr_to_jid(abbr: str) -> str:
    return metadata.lookup(abbr=abbr).jurisdiction_id



import typing
import re
import json
import click
import copy
import datetime
from pathlib import Path
from pydantic import BaseModel
from ... import metadata
from ..people import (
    get_new_filename,
    get_data_path,
    dump_obj,
    ocd_uuid,
    retire_person,
    retire_file,
)
from ...models.people import (
    Person,
    Role,
    Party,
    Office,
    Link,
    OtherName,
    OtherIdentifier,
)


def find_file(leg_id: str, *, state: str = "*") -> Path:
    if leg_id.startswith("ocd-person"):
        leg_id = leg_id.split("/")[1]
    assert len(leg_id) == 36

    if state == "*":
        filedir = get_data_path(".")
        files = list(filedir.glob(f"*/*/*{leg_id}.yml"))
    else:
        filedir = get_data_path(state)
        files = list(filedir.glob(f"*/*{leg_id}.yml"))

    if len(files) == 1:
        return files[0]
    elif len(files) > 1:
        raise ValueError(f"multiple files with same leg_id: {leg_id}")
    else:
        raise FileNotFoundError()


def collapse_duplicates(offices: list[Office]) -> list[Office]:
    output_offices = []
    used_offices: set[int] = set()

    for i, office1 in enumerate(offices):
        if i in used_offices:
            continue
        for j, office2 in enumerate(offices):
            if j in used_offices or i == j:
                continue
            # can one subsume the other?
            office1_wins = all(
                (
                    office1.voice == office2.voice or office2.voice == "",
                    office1.fax == office2.fax or office2.fax == "",
                    office1.address == office2.address or office2.address == "",
                )
            )
            office2_wins = all(
                (
                    office1.voice == office2.voice or office1.voice == "",
                    office1.fax == office2.fax or office1.fax == "",
                    office1.address == office2.address or office1.address == "",
                )
            )
            # if both are the same except the name, the one with the name wins
            if office1_wins and office2_wins:
                if office1.name:
                    office2_wins = False
                elif office2.name:
                    office1_wins = False

            if office1_wins:
                output_offices.append(office1)
                used_offices.update((i, j))
                break
            elif office2_wins:
                output_offices.append(office2)
                used_offices.update((i, j))
                break

        if i not in used_offices:
            output_offices.append(office1)
            used_offices.add(i)

    return output_offices


def merge_parties(old: list[Party], new: list[Party]) -> typing.Optional[list[Party]]:
    # only works for one new party for now, other cases aren't needed yet
    if len(new) > 1:
        raise ValueError(f"invalid new party config: {new}")
    if old[-1] == new[0]:
        return None
    retval = copy.deepcopy(old)

    if retval[-1].end_date == "":
        # these dates are imprecise for now
        retval[-1].end_date = datetime.date.today().strftime("%Y-%m-%d")
    retval.append(new[0])
    return retval


def merge_offices(
    old: list[Office], new: list[Office]
) -> typing.Optional[typing.List[Office]]:
    # figure out which office entries are which
    old_offices: dict[str, Office] = {}
    new_offices: dict[str, Office] = {}
    offices = []
    update = False

    for office in old:
        name = office.display_name
        if name not in old_offices:
            old_offices[name] = office
        else:
            raise NotImplementedError(f"extra old {name}")
    for office in new:
        name = office.display_name
        if name not in new_offices:
            new_offices[name] = office
        else:
            raise NotImplementedError(f"extra new {name}")

    for name in sorted(set(old_offices) | set(new_offices)):
        combined = update_office(old_offices.get(name), new_offices.get(name))
        offices.append(combined)
        if combined != old_offices.get(name):
            update = True

    # return all offices if there were any changes
    if update:
        return collapse_duplicates(offices)
    else:
        return None


def update_office(
    old_office: typing.Optional[Office],
    new_office: typing.Optional[Office],
) -> Office:
    """function returns a copy of old_office updated with values from new if applicable"""

    # if only one exists, return that one
    if not old_office and new_office:
        return new_office
    if not new_office and old_office:
        return old_office

    # combine the two
    if new_office and old_office:
        updated_office = old_office.copy()
        for field in updated_office.__fields__.keys():
            oldval = getattr(old_office, field)
            newval = getattr(new_office, field)
            if oldval != newval and newval:
                setattr(updated_office, field, newval)
    return updated_office


class Append:
    def __init__(self, key_name: str, list_item: typing.Any):
        self.key_name = key_name
        self.list_item = list_item

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Append)
            and self.key_name == other.key_name
            and self.list_item == other.list_item
        )

    def __str__(self) -> str:
        return f"{self.key_name}: append {dict(self.list_item)}"

    def __repr__(self) -> str:
        return f"Append({self.key_name}, {self.list_item})"


class Replace:
    def __init__(self, key_name: str, value_one: typing.Any, value_two: typing.Any):
        self.key_name = key_name
        self.value_one = value_one
        self.value_two = value_two

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Replace)
            and self.key_name == other.key_name
            and self.value_one == other.value_one
            and self.value_two == other.value_two
        )

    def __str__(self) -> str:
        return f"{self.key_name}: {self.value_one} => {self.value_two}"

    def __repr__(self) -> str:
        return f"Replace({self.key_name}, {self.value_one}, {self.value_two})"


class OfficesReplace(Replace):
    def __str__(self) -> str:
        def _fmt_cd(cd: Office) -> str:
            cd_str = f"{cd.display_name}"
            for key in ("address", "voice", "fax"):
                if val := getattr(cd, key):
                    cd_str += f" {key}={val}"
            return cd_str

        old = "\n\t".join(_fmt_cd(cd) for cd in self.value_one)
        new = "\n\t".join(_fmt_cd(cd) for cd in self.value_two)

        return f"{self.key_name} changed from: \n\t{old}\n  to\n\t{new}"


def compute_merge(
    obj1: BaseModel,
    obj2: BaseModel,
    prefix: str = "",
    keep_both_ids: bool = False,
    reset_offices: bool = False,
) -> list[typing.Union[Append, Replace]]:
    all_keys = obj1.__fields__.keys()
    changes: list[typing.Union[Append, Replace]] = []

    for key in all_keys:
        key_name = ".".join((prefix, key)) if prefix else key
        val1 = getattr(obj1, key)
        val2 = getattr(obj2, key)

        # special cases first
        if key == "id":
            if val1 != val2 and keep_both_ids:
                # old id stays as id: to keep things sane
                changes.append(
                    Append(
                        "other_identifiers",
                        OtherIdentifier(scheme="openstates", identifier=val2),
                    )
                )
        elif key == "name":
            if val1 != val2:
                # new name becomes name, but old name goes into other_names
                changes.append(Append("other_names", OtherName(name=val1)))
                changes.append(Replace("name", val1, val2))
        elif key == "party":
            changed = merge_parties(val1, val2)
            if changed:
                changes.append(Replace("party", val1, changed))
        elif key == "offices":
            if reset_offices:
                changed_offices = val2
            else:
                changed_offices = merge_offices(val1, val2)
            if changed_offices:
                changes.append(OfficesReplace("offices", val1 or [], changed_offices))
        elif isinstance(val1, list) or isinstance(val2, list):
            if val1 and not val2:
                continue
            elif val2 and not val1:
                changes.append(Replace(key_name, val1, val2))
            else:
                # both have elements, append new to old, leave old intact
                for item in typing.cast(list, val2):
                    if item not in typing.cast(list, val1):
                        changes.append(Append(key_name, item))
        elif isinstance(val1, BaseModel) or isinstance(val2, BaseModel):
            changes.extend(compute_merge(val1, val2, prefix=key_name))
        else:
            # if values both exist and differ, or val1 is empty, do a Replace
            if (val1 and val2 and val1 != val2) or (not val1 and val2):
                changes.append(Replace(key_name, val1, val2))

    return changes


def roles_equalish(role1: Role, role2: Role) -> bool:
    return (
        role1.type == role2.type
        and role1.jurisdiction == role2.jurisdiction
        and role1.district == role2.district
        and role1.end_date == role2.end_date
        and role1.end_reason == role2.end_reason
    )


def incoming_merge(
    abbr: str,
    existing_people: list[Person],
    new_people: list[Person],
    retirement: str,
    reset_offices: bool,
) -> list[tuple[Person, list[Person]]]:
    unmatched = []

    seats_for_district = {}
    state = metadata.lookup(abbr=abbr)
    for chamber in state.chambers:
        chtype = (
            "legislature"
            if chamber.chamber_type == "unicameral"
            else chamber.chamber_type
        )
        seats_for_district[chtype] = {
            district.name: district.num_seats for district in chamber.districts
        }

    # find candidate(s) for each new person
    for new in new_people:
        matched = False
        role_matches = []

        for existing in existing_people:
            name_match = new.name == existing.name
            role_match = False
            for role in existing.roles:
                if role.type == "mayor" or role.type == "governor":
                    continue
                seats = seats_for_district[role.type].get(
                    typing.cast(str, role.district), 1
                )
                # roles match if they are equal and there's only one seat, or
                # if there is already a name match on this legislator
                if roles_equalish(new.roles[0], role) and (seats == 1 or name_match):
                    role_match = True
                    # if they match without start date, copy the start date over so it isn't
                    # altered or otherwise removed in the merge
                    new.roles[0] = role
                    break
            if name_match or role_match:
                matched = interactive_merge(
                    abbr,
                    existing,
                    new,
                    name_match,
                    role_match,
                    retirement,
                    reset_offices,
                )

            if matched:
                break

            # if we haven't matched and this was a role match, save this for later
            if role_match:
                role_matches.append(existing)
        else:
            # not matched
            unmatched.append((new, role_matches))
            write_new_file(abbr, new, "legislature")

    return unmatched


def write_new_file(abbr: str, new: Person, _type: str) -> None:
    filedir = get_data_path(abbr)
    fname = get_new_filename(new.dict())
    newpath = filedir / _type / fname
    dump_obj(new, filename=newpath)
    click.secho(f"writing {new} to {newpath}", fg="yellow")


def retire(
    existing: Person, new: Person, retirement: typing.Optional[str] = None
) -> None:
    if not retirement:
        retirement = click.prompt("Enter retirement date YYYY-MM-DD")
    person, num = retire_person(existing, typing.cast(str, retirement))
    fname = find_file(existing.id)
    dump_obj(person, filename=fname)
    new_filename = retire_file(fname)
    click.secho(f"moved from {fname} to {new_filename}")


def interactive_merge(
    abbr: str,
    old: Person,
    new: Person,
    name_match: bool,
    role_match: bool,
    retirement: str,
    reset_offices: bool,
) -> bool:
    """
    returns True iff a merge was done
    """
    oldfname = find_file(old.id)
    # click.secho(" {} {}".format(oldfname, newfname), fg="yellow")

    # simulate difference
    changes = compute_merge(old, new, keep_both_ids=False, reset_offices=reset_offices)

    if not changes:
        click.secho(" perfect match", fg="green")
        return True

    for change in changes:
        if change.key_name == "name" or change.key_name == "roles":
            click.secho("    " + str(change), fg="red", bold=True)
        else:
            click.echo("    " + str(change))

    ch = "~"
    if name_match and role_match:
        choices = "m"
        # automatically pick merge
        ch = "m"
        # there is one very specific case that this fails in, if someone is beaten
        # by someone with the exact same name, that'll need to be caught manually
    elif name_match:
        choices = "m"
        text = "(m)erge?"
    elif role_match:
        choices = "mr"
        text = f"(m)erge? (r)etire {old.name}?"

    while ch not in (choices + "sa"):
        click.secho(text + " (s)kip? (a)bort?", bold=True)
        ch = click.getchar()

    if ch == "a":
        raise SystemExit(-1)
    elif ch == "m":
        merged = merge_people(
            old, new, keep_both_ids=False, reset_offices=reset_offices
        )
        dump_obj(merged, filename=oldfname)
        click.secho(" merged.", fg="green")
    elif ch == "r":
        # TODO: remove new file
        write_new_file(abbr, new, "legislature")
        retire(old, new, retirement)
    elif ch == "s":
        return False

    return True


def merge_people(
    old: Person, new: Person, keep_both_ids: bool = False, reset_offices: bool = False
) -> Person:
    """
    Function to merge two people objects.

    keep_both_ids
        Should be set to True iff people have been imported before.
        If we're dealing with an election, it should be set to false since the new ID
        hasn't been published anywhere yet.
    """
    changes = compute_merge(
        old,
        new,
        keep_both_ids=keep_both_ids,
        reset_offices=reset_offices,
    )

    for change in changes:
        if isinstance(change, Replace):
            keys = change.key_name.split(".")

            # recursively set the value based on dotted key
            temp_obj = old
            for key in keys[:-1]:
                temp_obj = getattr(temp_obj, key)
            setattr(temp_obj, keys[-1], change.value_two)
        if isinstance(change, Append):
            getattr(old, change.key_name).append(change.list_item)
    return old


PHONE_RE = re.compile(
    r"""^
                      \D*(1?)\D*                                # prefix
                      (\d{3})\D*(\d{3})\D*(\d{4}).*?             # main 10 digits
                      (?:(?:ext|Ext|EXT)\.?\s*\s*(\d{1,4}))?    # extension
                      $""",
    re.VERBOSE,
)


def reformat_phone_number(phone: str) -> str:
    match = PHONE_RE.match(phone)
    if match:
        groups = match.groups()

        ext = groups[-1]
        if ext:
            ext = f" ext. {ext}"
        else:
            ext = ""

        if not groups[0]:
            groups = groups[1:-1]
        else:
            groups = groups[:-1]
        return "-".join(groups) + ext
    else:
        return phone


def reformat_address(address: str) -> str:
    return re.sub(r"\s+", " ", re.sub(r"\s*\n\s*", ";", address))


def process_scrape_dir(input_dir: Path, jurisdiction_id: str) -> list[Person]:
    new_people = []

    # process people
    for filename in input_dir.glob("*.json"):
        with open(filename) as f:
            data = json.load(f)

        person = process_person(data, jurisdiction_id)
        new_people.append(person)

    return new_people


def process_office(office_data: dict) -> typing.Optional[Office]:
    voice = fax = address = ""
    if value := office_data["voice"]:
        voice = reformat_phone_number(value)
    if value := office_data["fax"]:
        fax = reformat_phone_number(value)
    if value := office_data["address"]:
        address = reformat_address(value)

    if voice or fax or address:
        return Office(
            classification=office_data["classification"],
            voice=voice,
            fax=fax,
            address=address,
        )
    else:
        return None


def process_person(data: dict, jurisdiction_id: str) -> Person:
    offices: list[Office] = []
    if office := data.pop("capitol_office"):
        cd = process_office(office)
        if cd:
            offices.append(cd)
    if office := data.pop("district_office"):
        cd = process_office(office)
        if cd:
            offices.append(cd)
    for office in data.pop("additional_offices"):
        cd = process_office(office)
        if cd:
            offices.append(cd)

    data.pop("state")
    chamber = data.pop("chamber")
    district = data.pop("district")

    result = Person(
        id=ocd_uuid("person"),
        roles=[
            Role(
                type=chamber,
                district=str(district),
                jurisdiction=jurisdiction_id,
            )
        ],
        party=[Party(name=data.pop("party"))],
        links=[Link(url=link["url"], note=link["note"]) for link in data.pop("links")],
        sources=[
            Link(url=link["url"], note=link["note"]) for link in data.pop("sources")
        ],
        offices=offices,
        **data,
    )

    return result



# Generated by Django 3.2 on 2021-04-21 15:52

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("data", "0033_auto_20210421_1538"),
    ]

    operations = [
        migrations.AddField(
            model_name="organization",
            name="other_names",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                blank=True, default=list
            ),
        ),
    ]



from ..models import State, Chamber, simple_numbered_districts

TX = State(
    name="Texas",
    abbr="TX",
    capital="Austin",
    capital_tz="America/Chicago",
    fips="48",
    unicameral=False,
    legislature_name="Texas Legislature",
    legislature_organization_id="ocd-organization/8ab77a54-0646-413c-a63a-dc85154282b7",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/4c8c2a9c-f33c-476d-bf81-266eb72193f0",
    division_id="ocd-division/country:us/state:tx",
    jurisdiction_id="ocd-jurisdiction/country:us/state:tx/government",
    url="https://capitol.texas.gov/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/d6189dbb-417e-429e-ae4b-2ee6747eddc0",
        num_seats=150,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:tx", "lower", 150
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/cabf1716-c572-406a-bfdd-1917c11ac629",
        num_seats=31,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:tx", "upper", 31
        ),
    ),
)



from ..models import State, Chamber, simple_numbered_districts

NE = State(
    name="Nebraska",
    abbr="NE",
    capital="Lincoln",
    capital_tz="America/Chicago",
    fips="31",
    unicameral=True,
    legislature_name="Nebraska Legislature",
    legislature_organization_id="ocd-organization/a0563aa3-1ab5-4e1e-afd1-643aca03b3b8",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/4ba67f40-0330-4d5a-ac31-8a939768d90b",
    division_id="ocd-division/country:us/state:ne",
    jurisdiction_id="ocd-jurisdiction/country:us/state:ne/government",
    url="http://nebraskalegislature.gov/",
    legislature=Chamber(
        chamber_type="unicameral",
        name="Nebraska Legislature",
        organization_id="ocd-organization/a0563aa3-1ab5-4e1e-afd1-643aca03b3b8",
        num_seats=49,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:ne", "legislature", 49
        ),
    ),
)



import click
import logging
import logging.config
from typing import Union
from openstates.utils import abbr_to_jid
from ..utils.django import init_django
from ..utils import transformers
from ..exceptions import InternalError
from .. import settings


# Attempt to fix bill identifiers in the DB that were NOT normalized when saved the first time
# non-normalized bill identifiers will never be matchable to a bill.identifier value
def fix_abnormal_related_bill_identifiers(jurisdiction_id: str) -> None:
    # import of model has to be after django_init
    from ..data.models import RelatedBill
    abnormal_unresolved_rb = RelatedBill.objects.filter(
        bill__legislative_session__jurisdiction_id=jurisdiction_id,
        related_bill=None,
    ).exclude(identifier__contains=' ')
    for rb in abnormal_unresolved_rb:
        new_identifier = transformers.fix_bill_id(rb.identifier)
        if new_identifier is not rb.identifier:
            # update this related bill row with normalized identifier
            rb.identifier = new_identifier
            rb.save()


@click.command(help="Resolve unresolved relationships between entities")
@click.argument("jurisdiction_abbreviation")
@click.option(
    "--log_level",
    help="Set the level of logging to output.",
    default="INFO"
)
@click.option(
    "--session",
    help="Session identifier, used to restrict resolution to within a specific session",
    default=None
)
def main(jurisdiction_abbreviation: str, log_level: str, session: Union[str, None]) -> None:
    # set up logging
    logger = logging.getLogger("openstates")
    handler_level = log_level
    settings.LOGGING["handlers"]["default"]["level"] = handler_level  # type: ignore
    logging.config.dictConfig(settings.LOGGING)

    # set up django for DB access
    # has to be done before any importer can be imported (?)
    init_django()
    from openstates.importers import resolve_related_bills

    logger.info(f"Beginning resolution of bill relationships for {jurisdiction_abbreviation}, session: {session}")
    jurisdiction_id = abbr_to_jid(jurisdiction_abbreviation)

    # Prep: resolve any non-normalized bill identifiers in related bill data
    # ie if RelatedBill has an identifier like "A1675" instead of "A 1675", then it can't be matched to a real bill
    # (this was a historical problem only fixed in mid 2024)
    fix_abnormal_related_bill_identifiers(jurisdiction_id)

    # Run the resolution logic
    try:
        resolve_related_bills(jurisdiction_id, session, logger)
    except InternalError as e:
        logger.error(f"Error during bill relationship resolution for {jurisdiction_abbreviation}: {e}")


if __name__ == "__main__":
    main()



from django.urls import reverse
from django.contrib import admin
from django.utils.safestring import mark_safe
from .. import models
from .base import (
    ModelAdmin,
    ReadOnlyTabularInline,
    IdentifierInline,
    OtherNameInline,
)


class PersonIdentifierInline(IdentifierInline):
    model = models.PersonIdentifier


class PersonNameInline(OtherNameInline):
    model = models.PersonName


class PersonOfficeInline(ReadOnlyTabularInline):
    readonly_fields = ("classification", "address", "voice", "fax", "name")
    model = models.PersonOffice


class PersonLinkInline(ReadOnlyTabularInline):
    readonly_fields = ("url", "note")
    model = models.PersonLink


class PersonSourceInline(ReadOnlyTabularInline):
    readonly_fields = ("url", "note")
    model = models.PersonSource


class MembershipInline(ReadOnlyTabularInline):
    model = models.Membership
    readonly_fields = ("organization", "person", "post", "role", "start_date")
    fields = ("id",) + readonly_fields + ("end_date",)
    exclude = ("id",)
    extra = 0
    can_delete = False
    ordering = ("end_date",)


# TODO field locking
@admin.register(models.Person)
class PersonAdmin(ModelAdmin):
    search_fields = ["name"]
    readonly_fields = ("id", "name", "extras")
    fields = (
        "name",
        "id",
        "image",
        ("birth_date", "death_date"),
        "gender",
        "biography",
        "extras",
    )
    ordering = ("name",)
    list_filter = ("memberships__organization__jurisdiction__name",)

    inlines = [
        PersonIdentifierInline,
        PersonNameInline,
        PersonOfficeInline,
        PersonLinkInline,
        PersonSourceInline,
        MembershipInline,
    ]

    def get_memberships(self, obj):
        memberships = obj.memberships.select_related("organization__jurisdiction")
        html = []
        SHOW_N = 5
        for memb in memberships[:SHOW_N]:
            org = memb.organization
            admin_url = reverse("admin:data_organization_change", args=(org.pk,))
            tmpl = '<a href="%s">%s%s</a>\n'
            html.append(
                tmpl
                % (
                    admin_url,
                    (
                        memb.organization.jurisdiction.name + ": "
                        if memb.organization.jurisdiction
                        else ""
                    ),
                    memb.organization.name,
                )
            )
        more = len(memberships) - SHOW_N
        if 0 < more:
            html.append("And %d more" % more)
        return mark_safe("<br/>".join(html))

    get_memberships.short_description = "Memberships"

    list_display = ("name", "id", "get_memberships")



from .common import sources, extras, fuzzy_datetime_blank
from ...data import common


schema = {
    "type": "object",
    "properties": {
        "identifier": {"type": "string"},
        "motion_text": {"type": "string", "minLength": 1},
        "motion_classification": {
            "items": {"type": "string", "enum": common.VOTE_CLASSIFICATIONS},
            "type": "array",
        },
        "start_date": fuzzy_datetime_blank,
        "end_date": fuzzy_datetime_blank,
        "result": {"type": "string", "enum": common.VOTE_RESULTS},
        "organization": {"type": ["string", "null"], "minLength": 1},
        "legislative_session": {"type": "string", "minLength": 1},
        "bill": {"type": ["string", "null"], "minLength": 1},
        "bill_action": {"type": ["string", "null"], "minLength": 1},
        "bill_identifier": {"type": "string"},
        "votes": {
            "items": {
                "type": "object",
                "properties": {
                    "option": {"type": "string", "enum": common.VOTE_OPTIONS},
                    "voter_name": {"type": "string", "minLength": 1},
                    "voter_id": {"type": "string", "minLength": 1},
                    "note": {"type": "string"},
                },
            }
        },
        "counts": {
            "items": {
                "properties": {
                    "option": {"type": "string", "enum": common.VOTE_OPTIONS},
                    "value": {"type": "integer", "minimum": 0},
                },
                "type": "object",
            }
        },
        "sources": sources,
        "extras": extras,
        "dedupe_key": {"type": ["string", "null"], "minLength": 1},
    },
}



"""
    Schema for event objects.
"""

from .common import (
    sources,
    extras,
    fuzzy_date_blank,
    fuzzy_datetime,
    fuzzy_datetime_blank,
)

media_schema = {
    "items": {
        "properties": {
            "name": {"type": "string", "minLength": 1},
            "type": {"type": "string", "minLength": 1},
            "date": fuzzy_date_blank,
            "offset": {"type": ["number", "null"]},
            "links": {
                "items": {
                    "properties": {
                        "media_type": {"type": "string"},
                        "url": {"type": "string", "format": "uri"},
                    },
                    "type": "object",
                },
                "type": "array",
            },
        },
        "type": "object",
    },
    "type": "array",
}

schema = {
    "properties": {
        "name": {"type": "string", "minLength": 1},
        "all_day": {"type": "boolean"},
        "start_date": fuzzy_datetime,
        "end_date": fuzzy_datetime_blank,
        "status": {
            "type": "string",
            "enum": ["cancelled", "tentative", "confirmed", "passed"],
        },
        "classification": {"type": "string", "minLength": 1},  # TODO: enum
        "description": {"type": "string"},
        "upstream_id": {"type": "string"},
        "location": {
            "type": "object",
            "properties": {
                "name": {"type": "string", "minLength": 1},
                "note": {"type": "string"},
                "url": {"type": ["string", "null"], "format": "uri"},
                "coordinates": {
                    "type": ["object", "null"],
                    "properties": {
                        "latitude": {"type": "string", "minLength": 1},
                        "longitude": {"type": "string", "minLength": 1},
                    },
                },
            },
        },
        "media": media_schema,
        "documents": {
            "items": {
                "properties": {
                    "note": {"type": "string", "minLength": 1},
                    "url": {"type": "string", "minLength": 1},
                    "media_type": {"type": "string", "minLength": 1},
                    "date": fuzzy_date_blank,
                },
                "type": "object",
            },
            "type": "array",
        },
        "links": {
            "items": {
                "properties": {
                    "note": {"type": "string"},
                    "url": {"format": "uri", "type": "string"},
                },
                "type": "object",
            },
            "type": "array",
        },
        "participants": {
            "items": {
                "properties": {
                    "name": {"type": "string", "minLength": 1},
                    "type": {"enum": ["organization", "person"], "type": "string"},
                    "note": {"type": "string", "minLength": 1},
                },
                "type": "object",
            },
            "type": "array",
        },
        "agenda": {
            "items": {
                "properties": {
                    "description": {"type": "string", "minLength": 1},
                    "classification": {
                        "items": {"type": "string", "minLength": 1},
                        "type": "array",
                    },
                    "order": {"type": "integer"},
                    "subjects": {
                        "items": {"type": "string", "minLength": 1},
                        "type": "array",
                    },
                    "media": media_schema,
                    "notes": {
                        "items": {"type": "string", "minLength": 1},
                        "type": "array",
                    },
                    "related_entities": {
                        "items": {
                            "properties": {
                                "entity_type": {"type": "string", "minLength": 1},
                                "name": {"type": "string", "minLength": 1},
                                "note": {"type": ["string", "null"], "minLength": 1},
                            },
                            "type": "object",
                        },
                        "minItems": 0,
                        "type": "array",
                    },
                },
                "type": "object",
            },
            "minItems": 0,
            "type": "array",
        },
        "sources": sources,
        "extras": extras,
        "dedupe_key": {"type": ["string", "null"], "minLength": 1},
    },
    "type": "object",
}



# Generated by Django 3.2.2 on 2021-08-02 17:52

import datetime
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("data", "0035_auto_20210422_1926"),
    ]

    operations = [
        migrations.AddField(
            model_name="jurisdiction",
            name="latest_bill_update",
            field=models.DateTimeField(default=datetime.date(2021, 1, 1)),
        ),
        migrations.AddField(
            model_name="jurisdiction",
            name="latest_people_update",
            field=models.DateTimeField(default=datetime.date(2021, 1, 1)),
        ),
    ]



from ..models import State, Chamber, District, simple_numbered_districts

SD = State(
    name="South Dakota",
    abbr="SD",
    capital="Pierre",
    capital_tz="America/Chicago",
    fips="46",
    unicameral=False,
    legislature_name="South Dakota State Legislature",
    legislature_organization_id="ocd-organization/7b114895-6d62-4b8d-b391-b170fa9011d2",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/c6e0fb4e-6cf7-4025-896c-f8cec8ad8cda",
    division_id="ocd-division/country:us/state:sd",
    jurisdiction_id="ocd-jurisdiction/country:us/state:sd/government",
    url="http://www.sdlegislature.gov/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/2432bd5a-3c3f-4784-b1f5-c6134171162b",
        num_seats=70,
        title="Representative",
        districts=[
            District("1", "lower", "ocd-division/country:us/state:sd/sldl:1", 2),
            District("2", "lower", "ocd-division/country:us/state:sd/sldl:2", 2),
            District("3", "lower", "ocd-division/country:us/state:sd/sldl:3", 2),
            District("4", "lower", "ocd-division/country:us/state:sd/sldl:4", 2),
            District("5", "lower", "ocd-division/country:us/state:sd/sldl:5", 2),
            District("6", "lower", "ocd-division/country:us/state:sd/sldl:6", 2),
            District("7", "lower", "ocd-division/country:us/state:sd/sldl:7", 2),
            District("8", "lower", "ocd-division/country:us/state:sd/sldl:8", 2),
            District("9", "lower", "ocd-division/country:us/state:sd/sldl:9", 2),
            District("10", "lower", "ocd-division/country:us/state:sd/sldl:10", 2),
            District("11", "lower", "ocd-division/country:us/state:sd/sldl:11", 2),
            District("12", "lower", "ocd-division/country:us/state:sd/sldl:12", 2),
            District("13", "lower", "ocd-division/country:us/state:sd/sldl:13", 2),
            District("14", "lower", "ocd-division/country:us/state:sd/sldl:14", 2),
            District("15", "lower", "ocd-division/country:us/state:sd/sldl:15", 2),
            District("16", "lower", "ocd-division/country:us/state:sd/sldl:16", 2),
            District("17", "lower", "ocd-division/country:us/state:sd/sldl:17", 2),
            District("18", "lower", "ocd-division/country:us/state:sd/sldl:18", 2),
            District("19", "lower", "ocd-division/country:us/state:sd/sldl:19", 2),
            District("20", "lower", "ocd-division/country:us/state:sd/sldl:20", 2),
            District("21", "lower", "ocd-division/country:us/state:sd/sldl:21", 2),
            District("22", "lower", "ocd-division/country:us/state:sd/sldl:22", 2),
            District("23", "lower", "ocd-division/country:us/state:sd/sldl:23", 2),
            District("24", "lower", "ocd-division/country:us/state:sd/sldl:24", 2),
            District("25", "lower", "ocd-division/country:us/state:sd/sldl:25", 2),
            District("26A", "lower", "ocd-division/country:us/state:sd/sldl:26a", 1),
            District("26B", "lower", "ocd-division/country:us/state:sd/sldl:26b", 1),
            District("27", "lower", "ocd-division/country:us/state:sd/sldl:27", 2),
            District("28A", "lower", "ocd-division/country:us/state:sd/sldl:28a", 1),
            District("28B", "lower", "ocd-division/country:us/state:sd/sldl:28b", 1),
            District("29", "lower", "ocd-division/country:us/state:sd/sldl:29", 2),
            District("30", "lower", "ocd-division/country:us/state:sd/sldl:30", 2),
            District("31", "lower", "ocd-division/country:us/state:sd/sldl:31", 2),
            District("32", "lower", "ocd-division/country:us/state:sd/sldl:32", 2),
            District("33", "lower", "ocd-division/country:us/state:sd/sldl:33", 2),
            District("34", "lower", "ocd-division/country:us/state:sd/sldl:34", 2),
            District("35", "lower", "ocd-division/country:us/state:sd/sldl:35", 2),
        ],
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/5a7260e0-bd1e-4f37-949f-ebb386e8bacc",
        num_seats=35,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:sd", "upper", 35
        ),
    ),
)



from ..models import State, Chamber, simple_numbered_districts

WY = State(
    name="Wyoming",
    abbr="WY",
    capital="Cheyenne",
    capital_tz="America/Denver",
    fips="56",
    unicameral=False,
    legislature_name="Wyoming State Legislature",
    legislature_organization_id="ocd-organization/b47d8d36-68d9-4bf0-807a-5b9c9ed2d99f",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/ef8bc870-8b19-4f99-a410-310d0ead92f4",
    division_id="ocd-division/country:us/state:wy",
    jurisdiction_id="ocd-jurisdiction/country:us/state:wy/government",
    url="http://legisweb.state.wy.us/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/683ca907-05d4-4c84-8d7b-f04f73e295ad",
        num_seats=62,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:wy", "lower", 62
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/6e0e804b-be2d-408b-9dc4-d8e3739915bd",
        num_seats=31,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:wy", "upper", 31
        ),
    ),
)



# Generated by Django 3.0.5 on 2020-07-21 16:04

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [("data", "0009_auto_20200429_0913")]

    operations = [
        migrations.AlterField(
            model_name="jurisdiction",
            name="division",
            field=models.ForeignKey(
                help_text="A link to a Division related to this Jurisdiction.",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="jurisdictions",
                to="data.Division",
            ),
        )
    ]



from ..models import State, Chamber, simple_numbered_districts

HI = State(
    name="Hawaii",
    abbr="HI",
    capital="Honolulu",
    capital_tz="Pacific/Honolulu",
    fips="15",
    unicameral=False,
    legislature_name="Hawaii State Legislature",
    legislature_organization_id="ocd-organization/c0ac26ed-60ea-42b7-bd02-20213257dbc5",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/8403b42b-dda0-5327-9ae1-599680374cc8",
    division_id="ocd-division/country:us/state:hi",
    jurisdiction_id="ocd-jurisdiction/country:us/state:hi/government",
    url="http://capitol.hawaii.gov",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/52cba4e0-aaeb-4d15-944b-089d559e97b6",
        num_seats=51,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:hi", "lower", 51
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/73058dca-37f9-4907-ac94-1f4ffd680586",
        num_seats=25,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:hi", "upper", 25
        ),
    ),
)



from ..models import State, Chamber, District, simple_numbered_districts

MN = State(
    name="Minnesota",
    abbr="MN",
    capital="Saint Paul",
    capital_tz="America/Chicago",
    fips="27",
    unicameral=False,
    legislature_name="Minnesota State Legislature",
    legislature_organization_id="ocd-organization/c5dddecb-4365-441f-bb5b-251dc6828d14",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/339ea74e-8ea4-4ce3-902d-6fd7a5bc571b",
    division_id="ocd-division/country:us/state:mn",
    jurisdiction_id="ocd-jurisdiction/country:us/state:mn/government",
    url="http://state.mn.us/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/e91db6f8-2232-49cd-91af-fdb5adb4ac3b",
        num_seats=134,
        title="Representative",
        districts=[
            District("1A", "lower", "ocd-division/country:us/state:mn/sldl:1a", 1),
            District("1B", "lower", "ocd-division/country:us/state:mn/sldl:1b", 1),
            District("2A", "lower", "ocd-division/country:us/state:mn/sldl:2a", 1),
            District("2B", "lower", "ocd-division/country:us/state:mn/sldl:2b", 1),
            District("3A", "lower", "ocd-division/country:us/state:mn/sldl:3a", 1),
            District("3B", "lower", "ocd-division/country:us/state:mn/sldl:3b", 1),
            District("4A", "lower", "ocd-division/country:us/state:mn/sldl:4a", 1),
            District("4B", "lower", "ocd-division/country:us/state:mn/sldl:4b", 1),
            District("5A", "lower", "ocd-division/country:us/state:mn/sldl:5a", 1),
            District("5B", "lower", "ocd-division/country:us/state:mn/sldl:5b", 1),
            District("6A", "lower", "ocd-division/country:us/state:mn/sldl:6a", 1),
            District("6B", "lower", "ocd-division/country:us/state:mn/sldl:6b", 1),
            District("7A", "lower", "ocd-division/country:us/state:mn/sldl:7a", 1),
            District("7B", "lower", "ocd-division/country:us/state:mn/sldl:7b", 1),
            District("8A", "lower", "ocd-division/country:us/state:mn/sldl:8a", 1),
            District("8B", "lower", "ocd-division/country:us/state:mn/sldl:8b", 1),
            District("9A", "lower", "ocd-division/country:us/state:mn/sldl:9a", 1),
            District("9B", "lower", "ocd-division/country:us/state:mn/sldl:9b", 1),
            District("10A", "lower", "ocd-division/country:us/state:mn/sldl:10a", 1),
            District("10B", "lower", "ocd-division/country:us/state:mn/sldl:10b", 1),
            District("11A", "lower", "ocd-division/country:us/state:mn/sldl:11a", 1),
            District("11B", "lower", "ocd-division/country:us/state:mn/sldl:11b", 1),
            District("12A", "lower", "ocd-division/country:us/state:mn/sldl:12a", 1),
            District("12B", "lower", "ocd-division/country:us/state:mn/sldl:12b", 1),
            District("13A", "lower", "ocd-division/country:us/state:mn/sldl:13a", 1),
            District("13B", "lower", "ocd-division/country:us/state:mn/sldl:13b", 1),
            District("14A", "lower", "ocd-division/country:us/state:mn/sldl:14a", 1),
            District("14B", "lower", "ocd-division/country:us/state:mn/sldl:14b", 1),
            District("15A", "lower", "ocd-division/country:us/state:mn/sldl:15a", 1),
            District("15B", "lower", "ocd-division/country:us/state:mn/sldl:15b", 1),
            District("16A", "lower", "ocd-division/country:us/state:mn/sldl:16a", 1),
            District("16B", "lower", "ocd-division/country:us/state:mn/sldl:16b", 1),
            District("17A", "lower", "ocd-division/country:us/state:mn/sldl:17a", 1),
            District("17B", "lower", "ocd-division/country:us/state:mn/sldl:17b", 1),
            District("18A", "lower", "ocd-division/country:us/state:mn/sldl:18a", 1),
            District("18B", "lower", "ocd-division/country:us/state:mn/sldl:18b", 1),
            District("19A", "lower", "ocd-division/country:us/state:mn/sldl:19a", 1),
            District("19B", "lower", "ocd-division/country:us/state:mn/sldl:19b", 1),
            District("20A", "lower", "ocd-division/country:us/state:mn/sldl:20a", 1),
            District("20B", "lower", "ocd-division/country:us/state:mn/sldl:20b", 1),
            District("21A", "lower", "ocd-division/country:us/state:mn/sldl:21a", 1),
            District("21B", "lower", "ocd-division/country:us/state:mn/sldl:21b", 1),
            District("22A", "lower", "ocd-division/country:us/state:mn/sldl:22a", 1),
            District("22B", "lower", "ocd-division/country:us/state:mn/sldl:22b", 1),
            District("23A", "lower", "ocd-division/country:us/state:mn/sldl:23a", 1),
            District("23B", "lower", "ocd-division/country:us/state:mn/sldl:23b", 1),
            District("24A", "lower", "ocd-division/country:us/state:mn/sldl:24a", 1),
            District("24B", "lower", "ocd-division/country:us/state:mn/sldl:24b", 1),
            District("25A", "lower", "ocd-division/country:us/state:mn/sldl:25a", 1),
            District("25B", "lower", "ocd-division/country:us/state:mn/sldl:25b", 1),
            District("26A", "lower", "ocd-division/country:us/state:mn/sldl:26a", 1),
            District("26B", "lower", "ocd-division/country:us/state:mn/sldl:26b", 1),
            District("27A", "lower", "ocd-division/country:us/state:mn/sldl:27a", 1),
            District("27B", "lower", "ocd-division/country:us/state:mn/sldl:27b", 1),
            District("28A", "lower", "ocd-division/country:us/state:mn/sldl:28a", 1),
            District("28B", "lower", "ocd-division/country:us/state:mn/sldl:28b", 1),
            District("29A", "lower", "ocd-division/country:us/state:mn/sldl:29a", 1),
            District("29B", "lower", "ocd-division/country:us/state:mn/sldl:29b", 1),
            District("30A", "lower", "ocd-division/country:us/state:mn/sldl:30a", 1),
            District("30B", "lower", "ocd-division/country:us/state:mn/sldl:30b", 1),
            District("31A", "lower", "ocd-division/country:us/state:mn/sldl:31a", 1),
            District("31B", "lower", "ocd-division/country:us/state:mn/sldl:31b", 1),
            District("32A", "lower", "ocd-division/country:us/state:mn/sldl:32a", 1),
            District("32B", "lower", "ocd-division/country:us/state:mn/sldl:32b", 1),
            District("33A", "lower", "ocd-division/country:us/state:mn/sldl:33a", 1),
            District("33B", "lower", "ocd-division/country:us/state:mn/sldl:33b", 1),
            District("34A", "lower", "ocd-division/country:us/state:mn/sldl:34a", 1),
            District("34B", "lower", "ocd-division/country:us/state:mn/sldl:34b", 1),
            District("35A", "lower", "ocd-division/country:us/state:mn/sldl:35a", 1),
            District("35B", "lower", "ocd-division/country:us/state:mn/sldl:35b", 1),
            District("36A", "lower", "ocd-division/country:us/state:mn/sldl:36a", 1),
            District("36B", "lower", "ocd-division/country:us/state:mn/sldl:36b", 1),
            District("37A", "lower", "ocd-division/country:us/state:mn/sldl:37a", 1),
            District("37B", "lower", "ocd-division/country:us/state:mn/sldl:37b", 1),
            District("38A", "lower", "ocd-division/country:us/state:mn/sldl:38a", 1),
            District("38B", "lower", "ocd-division/country:us/state:mn/sldl:38b", 1),
            District("39A", "lower", "ocd-division/country:us/state:mn/sldl:39a", 1),
            District("39B", "lower", "ocd-division/country:us/state:mn/sldl:39b", 1),
            District("40A", "lower", "ocd-division/country:us/state:mn/sldl:40a", 1),
            District("40B", "lower", "ocd-division/country:us/state:mn/sldl:40b", 1),
            District("41A", "lower", "ocd-division/country:us/state:mn/sldl:41a", 1),
            District("41B", "lower", "ocd-division/country:us/state:mn/sldl:41b", 1),
            District("42A", "lower", "ocd-division/country:us/state:mn/sldl:42a", 1),
            District("42B", "lower", "ocd-division/country:us/state:mn/sldl:42b", 1),
            District("43A", "lower", "ocd-division/country:us/state:mn/sldl:43a", 1),
            District("43B", "lower", "ocd-division/country:us/state:mn/sldl:43b", 1),
            District("44A", "lower", "ocd-division/country:us/state:mn/sldl:44a", 1),
            District("44B", "lower", "ocd-division/country:us/state:mn/sldl:44b", 1),
            District("45A", "lower", "ocd-division/country:us/state:mn/sldl:45a", 1),
            District("45B", "lower", "ocd-division/country:us/state:mn/sldl:45b", 1),
            District("46A", "lower", "ocd-division/country:us/state:mn/sldl:46a", 1),
            District("46B", "lower", "ocd-division/country:us/state:mn/sldl:46b", 1),
            District("47A", "lower", "ocd-division/country:us/state:mn/sldl:47a", 1),
            District("47B", "lower", "ocd-division/country:us/state:mn/sldl:47b", 1),
            District("48A", "lower", "ocd-division/country:us/state:mn/sldl:48a", 1),
            District("48B", "lower", "ocd-division/country:us/state:mn/sldl:48b", 1),
            District("49A", "lower", "ocd-division/country:us/state:mn/sldl:49a", 1),
            District("49B", "lower", "ocd-division/country:us/state:mn/sldl:49b", 1),
            District("50A", "lower", "ocd-division/country:us/state:mn/sldl:50a", 1),
            District("50B", "lower", "ocd-division/country:us/state:mn/sldl:50b", 1),
            District("51A", "lower", "ocd-division/country:us/state:mn/sldl:51a", 1),
            District("51B", "lower", "ocd-division/country:us/state:mn/sldl:51b", 1),
            District("52A", "lower", "ocd-division/country:us/state:mn/sldl:52a", 1),
            District("52B", "lower", "ocd-division/country:us/state:mn/sldl:52b", 1),
            District("53A", "lower", "ocd-division/country:us/state:mn/sldl:53a", 1),
            District("53B", "lower", "ocd-division/country:us/state:mn/sldl:53b", 1),
            District("54A", "lower", "ocd-division/country:us/state:mn/sldl:54a", 1),
            District("54B", "lower", "ocd-division/country:us/state:mn/sldl:54b", 1),
            District("55A", "lower", "ocd-division/country:us/state:mn/sldl:55a", 1),
            District("55B", "lower", "ocd-division/country:us/state:mn/sldl:55b", 1),
            District("56A", "lower", "ocd-division/country:us/state:mn/sldl:56a", 1),
            District("56B", "lower", "ocd-division/country:us/state:mn/sldl:56b", 1),
            District("57A", "lower", "ocd-division/country:us/state:mn/sldl:57a", 1),
            District("57B", "lower", "ocd-division/country:us/state:mn/sldl:57b", 1),
            District("58A", "lower", "ocd-division/country:us/state:mn/sldl:58a", 1),
            District("58B", "lower", "ocd-division/country:us/state:mn/sldl:58b", 1),
            District("59A", "lower", "ocd-division/country:us/state:mn/sldl:59a", 1),
            District("59B", "lower", "ocd-division/country:us/state:mn/sldl:59b", 1),
            District("60A", "lower", "ocd-division/country:us/state:mn/sldl:60a", 1),
            District("60B", "lower", "ocd-division/country:us/state:mn/sldl:60b", 1),
            District("61A", "lower", "ocd-division/country:us/state:mn/sldl:61a", 1),
            District("61B", "lower", "ocd-division/country:us/state:mn/sldl:61b", 1),
            District("62A", "lower", "ocd-division/country:us/state:mn/sldl:62a", 1),
            District("62B", "lower", "ocd-division/country:us/state:mn/sldl:62b", 1),
            District("63A", "lower", "ocd-division/country:us/state:mn/sldl:63a", 1),
            District("63B", "lower", "ocd-division/country:us/state:mn/sldl:63b", 1),
            District("64A", "lower", "ocd-division/country:us/state:mn/sldl:64a", 1),
            District("64B", "lower", "ocd-division/country:us/state:mn/sldl:64b", 1),
            District("65A", "lower", "ocd-division/country:us/state:mn/sldl:65a", 1),
            District("65B", "lower", "ocd-division/country:us/state:mn/sldl:65b", 1),
            District("66A", "lower", "ocd-division/country:us/state:mn/sldl:66a", 1),
            District("66B", "lower", "ocd-division/country:us/state:mn/sldl:66b", 1),
            District("67A", "lower", "ocd-division/country:us/state:mn/sldl:67a", 1),
            District("67B", "lower", "ocd-division/country:us/state:mn/sldl:67b", 1),
        ],
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/6a026144-758d-4d57-b856-9c60dce3c4b5",
        num_seats=67,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:mn", "upper", 67
        ),
    ),
)



# Generated by Django 3.0.5 on 2020-08-04 16:54

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("data", "0012_person_current_role")]

    operations = [migrations.RemoveField(model_name="person", name="current_state")]



#!/usr/bin/env python
import os
import typing
import sys
import csv
import math
import warnings
import click
import scrapelib
import time
from pathlib import Path
from django.contrib.postgres.search import SearchVector  # type: ignore
from django.db import transaction  # type: ignore
from django.db.models import Count  # type: ignore
from openstates.utils.django import init_django
from openstates.utils import jid_to_abbr, abbr_to_jid
from openstates.fulltext import (
    get_extract_func,
    DoNotDownload,
    CONVERSION_FUNCTIONS,
    Metadata,
)
from ..utils.instrument import Instrumentation

stats = Instrumentation()
# disable SSL validation and ignore warnings
scraper = scrapelib.Scraper(verify=False)
scraper.user_agent = "Mozilla"
warnings.filterwarnings("ignore", module="urllib3")


def get_raw_dir() -> Path:
    return Path(__file__).parent / ".." / "fulltext" / "raw"


MIMETYPES = {
    "application/pdf": "pdf",
    "text/html": "html",
    "application/msword": "doc",
    "application/rtf": "rtf",
    "application/vnd.openxmlformats-officedocument.wordprocessingml.document": "docx",
}


def _cleanup(text: str) -> str:
    # strip nulls
    return text.replace("\0", "")


def download(
    version: dict[str, str]
) -> tuple[typing.Optional[str], typing.Optional[bytes]]:
    abbr = jid_to_abbr(version["jurisdiction_id"])
    ext = MIMETYPES[version["media_type"]]
    filename = str(
        get_raw_dir()
        / f'{abbr}/{version["session"]}-{version["identifier"]}-{version["note"]}.{ext}'
    )
    filename.replace("#", "__")

    # FL "dh key too small" error due to bad Diffie Hellman key on the server side
    ciphers_list_addition = None
    if abbr == "fl":
        ciphers_list_addition = "HIGH:!DH:!aNULL"

    if not os.path.exists(filename):
        try:
            os.makedirs(os.path.dirname(filename))
        except OSError:
            pass
        try:
            _, resp = scraper.urlretrieve(
                version["url"], filename, ciphers_list_addition=ciphers_list_addition
            )
        except Exception:
            click.secho(f"could not fetch {version['url']}", fg="yellow")
            return None, None

        return filename, resp.content
    else:
        with open(filename, "rb") as f:
            return filename, f.read()


def extract_to_file(
    filename: str, data: bytes, version: Metadata
) -> tuple[typing.Union[None, str, typing.Type[DoNotDownload]], int]:
    text: typing.Optional[str]
    try:
        func = get_extract_func(version)
        if func == DoNotDownload:
            return DoNotDownload, 0
        else:
            text = func(data, version)
    except Exception as e:
        stats.write_stats(
            [
                {
                    "metric": "failed_text_extractions",
                    "fields": {"total": 1},
                    "tags": {"jurisdiction": version["jurisdiction_id"]},
                }
            ]
        )
        click.secho(f"exception processing {version['url']}: {e}", fg="red")
        text = None

    if not text:
        return None, 0

    text_filename = filename.replace("raw/", "text/") + ".txt"
    try:
        os.makedirs(os.path.dirname(text_filename))
    except OSError:
        pass
    with open(text_filename, "w") as f:
        f.write(text)

    return text_filename, len(text)


def update_bill(bill: typing.Any) -> int:
    from openstates.data.models import SearchableBill

    try:
        latest_version = bill.versions.order_by("-date", "-note").prefetch_related(
            "links"
        )[0]
        links = latest_version.links.all()
    except IndexError:
        links = []

    # check if there's an old entry and we can use it
    # if bill.searchable:
    #     if bill.searchable.version_id == latest_version.id and not bill.searchable.is_error:
    #         return      # nothing to do
    #     bill.searchable.delete()

    # FL "dh key too small" error due to bad Diffie Hellman key on the server side
    jurisdiction = bill.legislative_session.jurisdiction.name
    ciphers_list_addition = None
    if jurisdiction == "Florida":
        ciphers_list_addition = "HIGH:!DH:!aNULL"

    # iterate through versions until we extract some good text
    is_error = True
    raw_text = ""
    link = None
    for link in links:
        # TODO: if we need other exceptions, change this to a pluggable interface
        if (
            bill.legislative_session.jurisdiction_id
            == "ocd-jurisdiction/country:us/state:ca/government"
        ):
            # move CA query string onto a docs-proxy query string for working PDF extraction
            old_url = link.url
            new_url = "http://docs-proxy.openstates.org/ca?" + link.url.split("?")[1]
            print(f"{old_url} => {new_url}")
            link.url = new_url
        metadata: Metadata = {
            "url": link.url,
            "media_type": link.media_type,
            "title": bill.title,
            "jurisdiction_id": bill.legislative_session.jurisdiction_id,
        }
        func = get_extract_func(metadata)
        if func == DoNotDownload:
            continue
        try:
            data = scraper.request(
                "GET",
                link.url,
                allow_redirects=True,
                ciphers_list_addition=ciphers_list_addition,
            ).content
        except Exception:
            continue
        try:
            raw_text = func(data, metadata)
        except Exception as e:
            click.secho(f"exception processing {metadata['url']}: {e}", fg="red")

        # TODO: clean up whitespace
        raw_text = _cleanup(raw_text)

        if raw_text:
            is_error = False
            break

    sb = SearchableBill.objects.create(
        bill=bill,
        version_link=link,
        all_titles=bill.title,  # TODO: add other titles
        raw_text=raw_text,
        is_error=is_error,
        search_vector="",
    )
    return sb.id


@click.group()
def main() -> None:
    pass


def _resample(state: str, n: int = 50) -> None:
    """
    Grab new versions for a state from the database.
    """
    init_django()
    from openstates.data.models import BillVersion

    versions = BillVersion.objects.filter(
        bill__legislative_session__jurisdiction_id=abbr_to_jid(state)
    ).order_by("?")[:n]

    count = 0
    fieldnames = [
        "id",
        "session",
        "identifier",
        "title",
        "jurisdiction_id",
        "media_type",
        "url",
        "note",
    ]

    with open(get_raw_dir() / f"{state}.csv", "w") as outf:
        out = csv.DictWriter(outf, fieldnames=fieldnames)
        out.writeheader()
        for v in versions:
            for link in v.links.all():
                out.writerow(
                    {
                        "id": v.id,
                        "session": v.bill.legislative_session.identifier,
                        "jurisdiction_id": v.bill.legislative_session.jurisdiction_id,
                        "identifier": v.bill.identifier,
                        "title": v.bill.title,
                        "url": link.url,
                        "media_type": link.media_type,
                        "note": v.note,
                    }
                )
                count += 1
    click.secho(f"wrote new sample csv with {count} records")


@main.command(help="obtain a sample of bills to extract text from")
@click.argument("state")
@click.option("--resample/--no-resample", default=False)
@click.option("--quiet/--no-quiet", default=False)
def sample(state: str, resample: bool, quiet: bool) -> int:
    if resample:
        _resample(state)
    count = missing = empty = skipped = 0
    with open(get_raw_dir() / f"{state}.csv") as f:
        for version in csv.DictReader(f):
            count += 1
            filename, data = download(version)
            if not filename or not data:
                missing += 1
                continue
            text_filename, n_bytes = extract_to_file(
                filename, data, typing.cast(Metadata, version)
            )
            if text_filename == DoNotDownload:
                skipped += 1
            elif not n_bytes:
                empty += 1
            if not quiet:
                click.secho(f"{filename} => {text_filename} ({n_bytes} bytes)")
    # decide and print result
    status = "green"
    if empty or missing:  # arbitrary threshold for now
        status = "red"
    click.secho(
        f"{state}: processed {count}, {skipped} skipped, {missing} missing, {empty} empty",
        fg=status,
    )
    if status == "red":
        return 1
    return 0


@main.command(help="run sample on all states, used for CI")
@click.pass_context
def test(ctx: typing.Any) -> None:
    failures = 0
    states = sorted(CONVERSION_FUNCTIONS.keys())
    click.secho(f"testing {len(states)} states...", fg="white")
    for state in states:
        failures += ctx.invoke(sample, state=state, quiet=True)
    sys.exit(failures)


@main.command(help="print a status table showing the current condition of states")
def status() -> None:
    init_django()
    from openstates.data.models import Bill

    states = sorted(CONVERSION_FUNCTIONS.keys())
    click.secho("state |  bills  | missing | errors ", fg="white")
    click.secho("===================================", fg="white")
    for state in states:
        all_bills = Bill.objects.filter(
            legislative_session__jurisdiction_id=abbr_to_jid(state)
        )
        missing_search = all_bills.filter(searchable__isnull=True).count()
        errors = all_bills.filter(searchable__is_error=True).count()
        all_bills = all_bills.count()

        errcolor = mscolor = "green"
        if missing_search > 0:
            missing_search = math.ceil(missing_search / all_bills * 100)
            mscolor = "yellow"
        if missing_search > 1:
            mscolor = "red"
        if errors > 0:
            errcolor = "yellow"
            errors = math.ceil(errors / all_bills * 100)
        if errors > 5:
            errcolor = "red"

        click.echo(
            f"{state:5} | {all_bills:7} | "
            + click.style(f"{missing_search:6}%", fg=mscolor)
            + " | "
            + click.style(f"{errors:6}%", fg=errcolor)
        )


@main.command(help="rebuild the search index objects for a given state")
@click.argument("state")
@click.option("--session", default=None)
def reindex_state(state: str, session: str = None) -> None:
    init_django()
    from openstates.data.models import SearchableBill

    if session:
        bills = SearchableBill.objects.filter(
            bill__legislative_session__jurisdiction_id=abbr_to_jid(state),
            bill__legislative_session__identifier=session,
        )
    else:
        bills = SearchableBill.objects.filter(
            bill__legislative_session__jurisdiction_id=abbr_to_jid(state)
        )

    ids = list(bills.values_list("id", flat=True))
    print(f"reindexing {len(ids)} bills for state")
    reindex(ids)


@main.command(help="update the saved bill text in the database")
@click.argument("state")
@click.option("-n", default=None)
@click.option("--clear-errors/--no-clear-errors", default=False)
@click.option("--checkpoint", default=500)
@click.option("--session", default=None)
def update(
    state: str, n: int, clear_errors: bool, checkpoint: int, session: str = None
) -> None:
    init_django()
    from openstates.data.models import Bill, SearchableBill

    # print status within checkpoints
    status_num = checkpoint / 5

    stats.write_stats(
        [
            {
                "metric": "text_extraction_runs",
                "fields": {"total": 1},
                "tags": {"jurisdiction": state},
            }
        ]
    )

    if state == "all":
        all_bills = Bill.objects.all()
    elif session:
        all_bills = Bill.objects.filter(
            legislative_session__jurisdiction_id=abbr_to_jid(state),
            legislative_session__identifier=session,
        )
    else:
        all_bills = Bill.objects.filter(
            legislative_session__jurisdiction_id=abbr_to_jid(state)
        )

    if clear_errors:
        if state == "all":
            print("--clear-errors only works with specific states, not all")
            return
        errs = SearchableBill.objects.filter(bill__in=all_bills, is_error=True)
        print(f"clearing {len(errs)} errors")
        errs.delete()

    missing_search = all_bills.filter(searchable__isnull=True)
    if state == "all":
        MAX_UPDATE = 1000
        aggregates = missing_search.values(
            "legislative_session__jurisdiction__name"
        ).annotate(count=Count("id"))
        for agg in aggregates:
            state_name = agg["legislative_session__jurisdiction__name"]
            if agg["count"] > MAX_UPDATE:
                click.secho(
                    f"Too many bills to update for {state_name}: {agg['count']}, skipping",
                    fg="red",
                )
                missing_search = missing_search.exclude(
                    legislative_session__jurisdiction__name=state_name
                )
        print(f"{len(missing_search)} missing, updating")
    else:
        print(
            f"{state}: {len(all_bills)} bills, {len(missing_search)} without search results"
        )
    stats.write_stats(
        [
            {
                "metric": "text_extraction_missing",
                "fields": {"vectors": len(missing_search)},
                "tags": {"jurisdiction": state},
            }
        ]
    )

    if n:
        missing_search = missing_search[: int(n)]
    else:
        n = len(missing_search)

    ids_to_update = []
    updated_count = 0

    # going to manage our own transactions here so we can save in chunks
    transaction.set_autocommit(False)

    for b in missing_search:
        ids_to_update.append(update_bill(b))
        updated_count += 1
        if updated_count % status_num == 0:
            print(f"{state}: updated {updated_count} out of {n}")
        if updated_count % checkpoint == 0:
            reindex(ids_to_update)
            transaction.commit()
            ids_to_update = []

    stats.write_stats(
        [
            {
                "metric": "text_extraction",
                "fields": {"updates": len(ids_to_update)},
                "tags": {"jurisdiction": state},
            }
        ]
    )
    # be sure to reindex final set
    reindex(ids_to_update)
    transaction.commit()
    transaction.set_autocommit(True)
    stats.write_stats(
        [
            {
                "metric": "last_text_extract",
                "fields": {"time": int(time.time())},
                "tags": {"jurisdiction": state},
            }
        ]
    )
    stats.close()


def reindex(ids_to_update: list[int]) -> None:
    from openstates.data.models import SearchableBill

    print(f"updating {len(ids_to_update)} search vectors")
    res = SearchableBill.objects.filter(id__in=ids_to_update).update(
        search_vector=(
            SearchVector("all_titles", weight="A", config="english")
            + SearchVector("raw_text", weight="B", config="english")
        )
    )
    print(f"updated {res}")


if __name__ == "__main__":
    main()



from ..models import State, Chamber, simple_numbered_districts

IA = State(
    name="Iowa",
    abbr="IA",
    capital="Des Moines",
    capital_tz="America/Chicago",
    fips="19",
    unicameral=False,
    legislature_name="Iowa General Assembly",
    legislature_organization_id="ocd-organization/643af5ff-e840-4ff1-95a8-e47a40dd0730",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/8d31cf5b-f2c1-5c52-b148-4e90cf8161b8",
    division_id="ocd-division/country:us/state:ia",
    jurisdiction_id="ocd-jurisdiction/country:us/state:ia/government",
    url="https://www.legis.iowa.gov/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/bd11534f-0c3b-4321-8d04-6a42f212bf77",
        num_seats=100,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:ia", "lower", 100
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/a9ad8273-4a86-4544-9bed-09ccb48ebabc",
        num_seats=50,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:ia", "upper", 50
        ),
    ),
)



# Generated by Django 3.0.5 on 2020-09-23 13:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("data", "0025_auto_20200914_1216"),
    ]

    operations = [
        migrations.AddField(
            model_name="billdocument",
            name="classification",
            field=models.CharField(
                blank=True,
                choices=[
                    ("", "Unknown"),
                    ("fiscal-note", "Fiscal Note"),
                    ("committee-report", "Committee Report"),
                    ("summary", "Summary"),
                    ("digest", "Digest"),
                    ("veto-message", "Veto Message"),
                    ("analysis", "Analysis"),
                    ("law", "Law"),
                ],
                max_length=100,
            ),
        ),
        migrations.AddField(
            model_name="billversion",
            name="classification",
            field=models.CharField(
                blank=True,
                choices=[
                    ("", "Unknown"),
                    ("filed", "Filed"),
                    ("introduced", "Introduced"),
                    ("amendment", "Amendment"),
                    ("substituted", "Substituted"),
                    ("enrolled", "Enrolled"),
                    ("became-law", "Became Law"),
                ],
                max_length=100,
            ),
        ),
        migrations.AlterField(
            model_name="jurisdiction",
            name="classification",
            field=models.CharField(
                choices=[("state", "State"), ("municipality", "Municipality")],
                db_index=True,
                default="state",
                help_text="The type of Jurisdiction being defined.",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="organization",
            name="classification",
            field=models.CharField(
                blank=True,
                choices=[
                    ("legislature", "Legislature"),
                    ("executive", "Executive"),
                    ("upper", "Upper Chamber"),
                    ("lower", "Lower Chamber"),
                    ("party", "Party"),
                    ("committee", "Committee"),
                    ("government", "Government"),
                ],
                help_text="The type of Organization being defined.",
                max_length=100,
            ),
        ),
    ]



from ..models import State, Chamber, simple_numbered_districts

AS = State(
    name="American Samoa",
    abbr="AS",
    capital="Pago Pago",
    capital_tz="Pacific/Samoa",
    fips="60",
    unicameral=False,
    url="https://www.americansamoa.gov",
    division_id="ocd-division/country:us/territory:as",
    jurisdiction_id="ocd-jurisdiction/country:us/territory:as/government",
    legislature_name="American Samoa Fono",
    legislature_organization_id="ocd-organization/739d2a37-30aa-46ca-a0ce-41bdd9f39f51",
    executive_name="Governor of American Samoa",
    executive_organization_id="ocd-organization/1b28142e-dcb4-457e-b17a-11298ae7ccf8",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/f6355e28-9d90-4fa2-8e58-8429a5d0e2a3",
        num_seats=21,
        title="Legislators",
        districts=simple_numbered_districts(
            "ocd-division/country:us/territory:as", "lower", 21
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/9a527967-b7f1-4186-a635-611a1f29dd1d",
        num_seats=18,
        title="Legislators",
        districts=simple_numbered_districts(
            "ocd-division/country:us/territory:as", "upper", 18
        ),
    ),
)



from ..models import State, Chamber, simple_numbered_districts

NV = State(
    name="Nevada",
    abbr="NV",
    capital="Carson City",
    capital_tz="America/Los_Angeles",
    fips="32",
    unicameral=False,
    legislature_name="Nevada Legislature",
    legislature_organization_id="ocd-organization/770c2aae-0332-4f24-a8e0-e0bf666dc992",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/0328f08e-ceb3-4d5c-a8d3-a73a3443e02e",
    division_id="ocd-division/country:us/state:nv",
    jurisdiction_id="ocd-jurisdiction/country:us/state:nv/government",
    url="http://www.leg.state.nv.us/",
    lower=Chamber(
        chamber_type="lower",
        name="Assembly",
        organization_id="ocd-organization/258cdef7-dac8-4ec1-8f74-00a70ee10269",
        num_seats=42,
        title="Assembly Member",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:nv", "lower", 42
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/c4400c7b-157d-442a-b688-543ffb989967",
        num_seats=21,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:nv", "upper", 21
        ),
    ),
)



# Generated by Django 3.2.2 on 2021-10-07 20:29

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ("data", "0040_auto_20210920_2347"),
    ]

    operations = [
        migrations.CreateModel(
            name="PersonOffice",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "classification",
                    models.CharField(
                        choices=[
                            ("district", "District Office"),
                            ("capitol", "Capitol Office"),
                            ("primary", "Primary Office"),
                        ],
                        max_length=20,
                    ),
                ),
                ("address", models.CharField(blank=True, default="", max_length=300)),
                ("voice", models.CharField(blank=True, default="", max_length=30)),
                ("fax", models.CharField(blank=True, default="", max_length=30)),
                ("name", models.CharField(blank=True, default="", max_length=200)),
                (
                    "person",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="offices",
                        to="data.person",
                    ),
                ),
            ],
            options={
                "db_table": "openstates_personoffice",
            },
        ),
    ]



from ..models import State, Chamber, simple_numbered_districts

WA = State(
    name="Washington",
    abbr="WA",
    capital="Olympia",
    capital_tz="America/Los_Angeles",
    fips="53",
    unicameral=False,
    legislature_name="Washington State Legislature",
    legislature_organization_id="ocd-organization/a7819d04-7b5e-4d2e-b47f-5dca9f09fc81",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/cb29ec2f-1ee0-50b9-8394-f43504579f65",
    division_id="ocd-division/country:us/state:wa",
    jurisdiction_id="ocd-jurisdiction/country:us/state:wa/government",
    url="http://www.leg.wa.gov",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/7417a716-eba8-4437-a6e9-df14694bcb79",
        num_seats=98,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:wa", "lower", 49, num_seats=2
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/ddae63d3-9f3e-4c46-a00f-703fe21b54d1",
        num_seats=49,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:wa", "upper", 49
        ),
    ),
)



from ..models import State, Chamber, District, simple_numbered_districts

ID = State(
    name="Idaho",
    abbr="ID",
    capital="Boise",
    capital_tz="America/Denver",
    fips="16",
    unicameral=False,
    legislature_name="Idaho State Legislature",
    legislature_organization_id="ocd-organization/654d4d38-e413-4307-a420-6f1ee2594673",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/9bbaab73-ae4e-5d25-9f32-c3af64a77829",
    division_id="ocd-division/country:us/state:id",
    jurisdiction_id="ocd-jurisdiction/country:us/state:id/government",
    url="http://www.legislature.idaho.gov",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/7a59916c-201a-49b2-9288-776ffea9f6cb",
        num_seats=70,
        title="Representative",
        districts=[
            District("1A", "lower", "ocd-division/country:us/state:id/sldl:1a", 1),
            District("1B", "lower", "ocd-division/country:us/state:id/sldl:1b", 1),
            District("2A", "lower", "ocd-division/country:us/state:id/sldl:2a", 1),
            District("2B", "lower", "ocd-division/country:us/state:id/sldl:2b", 1),
            District("3A", "lower", "ocd-division/country:us/state:id/sldl:3a", 1),
            District("3B", "lower", "ocd-division/country:us/state:id/sldl:3b", 1),
            District("4A", "lower", "ocd-division/country:us/state:id/sldl:4a", 1),
            District("4B", "lower", "ocd-division/country:us/state:id/sldl:4b", 1),
            District("5A", "lower", "ocd-division/country:us/state:id/sldl:5a", 1),
            District("5B", "lower", "ocd-division/country:us/state:id/sldl:5b", 1),
            District("6A", "lower", "ocd-division/country:us/state:id/sldl:6a", 1),
            District("6B", "lower", "ocd-division/country:us/state:id/sldl:6b", 1),
            District("7A", "lower", "ocd-division/country:us/state:id/sldl:7a", 1),
            District("7B", "lower", "ocd-division/country:us/state:id/sldl:7b", 1),
            District("8A", "lower", "ocd-division/country:us/state:id/sldl:8a", 1),
            District("8B", "lower", "ocd-division/country:us/state:id/sldl:8b", 1),
            District("9A", "lower", "ocd-division/country:us/state:id/sldl:9a", 1),
            District("9B", "lower", "ocd-division/country:us/state:id/sldl:9b", 1),
            District("10A", "lower", "ocd-division/country:us/state:id/sldl:10a", 1),
            District("10B", "lower", "ocd-division/country:us/state:id/sldl:10b", 1),
            District("11A", "lower", "ocd-division/country:us/state:id/sldl:11a", 1),
            District("11B", "lower", "ocd-division/country:us/state:id/sldl:11b", 1),
            District("12A", "lower", "ocd-division/country:us/state:id/sldl:12a", 1),
            District("12B", "lower", "ocd-division/country:us/state:id/sldl:12b", 1),
            District("13A", "lower", "ocd-division/country:us/state:id/sldl:13a", 1),
            District("13B", "lower", "ocd-division/country:us/state:id/sldl:13b", 1),
            District("14A", "lower", "ocd-division/country:us/state:id/sldl:14a", 1),
            District("14B", "lower", "ocd-division/country:us/state:id/sldl:14b", 1),
            District("15A", "lower", "ocd-division/country:us/state:id/sldl:15a", 1),
            District("15B", "lower", "ocd-division/country:us/state:id/sldl:15b", 1),
            District("16A", "lower", "ocd-division/country:us/state:id/sldl:16a", 1),
            District("16B", "lower", "ocd-division/country:us/state:id/sldl:16b", 1),
            District("17A", "lower", "ocd-division/country:us/state:id/sldl:17a", 1),
            District("17B", "lower", "ocd-division/country:us/state:id/sldl:17b", 1),
            District("18A", "lower", "ocd-division/country:us/state:id/sldl:18a", 1),
            District("18B", "lower", "ocd-division/country:us/state:id/sldl:18b", 1),
            District("19A", "lower", "ocd-division/country:us/state:id/sldl:19a", 1),
            District("19B", "lower", "ocd-division/country:us/state:id/sldl:19b", 1),
            District("20A", "lower", "ocd-division/country:us/state:id/sldl:20a", 1),
            District("20B", "lower", "ocd-division/country:us/state:id/sldl:20b", 1),
            District("21A", "lower", "ocd-division/country:us/state:id/sldl:21a", 1),
            District("21B", "lower", "ocd-division/country:us/state:id/sldl:21b", 1),
            District("22A", "lower", "ocd-division/country:us/state:id/sldl:22a", 1),
            District("22B", "lower", "ocd-division/country:us/state:id/sldl:22b", 1),
            District("23A", "lower", "ocd-division/country:us/state:id/sldl:23a", 1),
            District("23B", "lower", "ocd-division/country:us/state:id/sldl:23b", 1),
            District("24A", "lower", "ocd-division/country:us/state:id/sldl:24a", 1),
            District("24B", "lower", "ocd-division/country:us/state:id/sldl:24b", 1),
            District("25A", "lower", "ocd-division/country:us/state:id/sldl:25a", 1),
            District("25B", "lower", "ocd-division/country:us/state:id/sldl:25b", 1),
            District("26A", "lower", "ocd-division/country:us/state:id/sldl:26a", 1),
            District("26B", "lower", "ocd-division/country:us/state:id/sldl:26b", 1),
            District("27A", "lower", "ocd-division/country:us/state:id/sldl:27a", 1),
            District("27B", "lower", "ocd-division/country:us/state:id/sldl:27b", 1),
            District("28A", "lower", "ocd-division/country:us/state:id/sldl:28a", 1),
            District("28B", "lower", "ocd-division/country:us/state:id/sldl:28b", 1),
            District("29A", "lower", "ocd-division/country:us/state:id/sldl:29a", 1),
            District("29B", "lower", "ocd-division/country:us/state:id/sldl:29b", 1),
            District("30A", "lower", "ocd-division/country:us/state:id/sldl:30a", 1),
            District("30B", "lower", "ocd-division/country:us/state:id/sldl:30b", 1),
            District("31A", "lower", "ocd-division/country:us/state:id/sldl:31a", 1),
            District("31B", "lower", "ocd-division/country:us/state:id/sldl:31b", 1),
            District("32A", "lower", "ocd-division/country:us/state:id/sldl:32a", 1),
            District("32B", "lower", "ocd-division/country:us/state:id/sldl:32b", 1),
            District("33A", "lower", "ocd-division/country:us/state:id/sldl:33a", 1),
            District("33B", "lower", "ocd-division/country:us/state:id/sldl:33b", 1),
            District("34A", "lower", "ocd-division/country:us/state:id/sldl:34a", 1),
            District("34B", "lower", "ocd-division/country:us/state:id/sldl:34b", 1),
            District("35A", "lower", "ocd-division/country:us/state:id/sldl:35a", 1),
            District("35B", "lower", "ocd-division/country:us/state:id/sldl:35b", 1),
        ],
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/e8c4c8ae-7ba0-4116-9865-5d34fcc6e681",
        num_seats=35,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:id", "upper", 35
        ),
    ),
)
